from requests import Session

from octodns import __version__ as octodns_version

from .exceptions import ApiException


class DNSClient:
    API_URL = 'https://api.selectel.ru/domains/v2'
    _PAGINATION_LIMIT = 50

    _zone_path = "/zones"
    __rrsets_path = "/zones/{}/rrset"
    __rrsets_path_specific = "/zones/{}/rrset/{}"

    def __init__(self, library_version: str, openstack_token: str):
        self._sess = Session()
        self._sess.headers.update(
            {
                'X-Auth-Token': openstack_token,
                'Content-Type': 'application/json',
                'User-Agent': f'octodns/{octodns_version} octodns-selectel/{library_version}',
            }
        )

    @classmethod
    def _rrset_path(cls, zone_id):
        return cls.__rrsets_path.format(zone_id)

    @classmethod
    def _rrset_path_specific(cls, zone_id, rrset_id):
        return cls.__rrsets_path_specific.format(zone_id, rrset_id)

    def _request(self, method, path, params=None, data=None):
        url = f'{self.API_URL}{path}'
        resp = self._sess.request(method, url, params, json=data)
        try:
            resp_json = resp.json()
        except ValueError:
            resp_json = {}
        if resp.status_code in {200, 201, 204}:
            return resp_json
        elif resp.status_code in {400, 422}:
            raise ApiException(
                f'Bad request. Description: {resp_json.get("description", "Invalid payload")}.'
            )
        elif resp.status_code == 401:
            raise ApiException('Authorization failed. Invalid or empty token.')
        elif resp.status_code == 404:
            raise ApiException(
                'Resource not found: '
                f'{resp_json.get("error", "invalid path")}.'
            )
        elif resp.status_code == 409:
            raise ApiException(
                f'Conflict: {resp_json.get("error", "resource maybe already created")}.'
            )
        else:
            raise ApiException('Internal server error.')

    def _request_all_entities(self, path, offset=0):
        items = []
        resp = self._request(
            "GET", path, dict(limit=self._PAGINATION_LIMIT, offset=offset)
        )
        items.extend(resp["result"])
        if next_offset := resp["next_offset"]:
            items.extend(self._request_all_entities(path, offset=next_offset))
        return items

    def list_zones(self):
        return self._request_all_entities(self._zone_path)

    def create_zone(self, name):
        return self._request('POST', self._zone_path, data=dict(name=name))

    def list_rrsets(self, zone_id):
        path = self._rrset_path(zone_id)
        return self._request_all_entities(path)

    def create_rrset(self, zone_id, data):
        path = self._rrset_path(zone_id)
        return self._request('POST', path, data=data)

    def update_rrset(self, zone_id, rrset_id, data):
        path = self._rrset_path_specific(zone_id, rrset_id)
        return self._request('PATCH', path, data=data)

    def delete_rrset(self, zone_id, rrset_id):
        path = self._rrset_path_specific(zone_id, rrset_id)
        return self._request('DELETE', path)



from .v1.provider import SelectelProvider as SelectelProviderLegacy
from .v2.provider import SelectelProvider as SelectelProvider
from .version import __VERSION__, __version__

__all__ = [SelectelProviderLegacy, SelectelProvider]

# quell warnings
__VERSION__
__version__



from string import Template

from .exceptions import SelectelException


def to_selectel_rrset(record):
    rrset = dict(name=record.fqdn, ttl=record.ttl, type=record._type)
    rrset_records = []
    content_mx_tmpl = Template("$preference $exchange")
    content_srv_tmpl = Template("$priority $weight $port $target")
    content_sshfp_tmpl = Template("$algorithm $fingerprint_type $fingerprint")
    if record._type in {"A", "AAAA", "NS"}:
        rrset_records = list(
            map(lambda value: {'content': value}, record.values)
        )
    elif record._type in {"CNAME", "ALIAS"}:
        rrset_records = [{'content': record.value}]
    elif record._type == "TXT":
        rrset_records = [
            dict(content=f'\"{value}\"') for value in record.values
        ]
    elif record._type == "MX":
        rrset_records = list(
            map(
                lambda value: {
                    'content': content_mx_tmpl.substitute(
                        preference=value.preference, exchange=value.exchange
                    )
                },
                record.values,
            )
        )
    elif record._type == "SRV":
        rrset_records = list(
            map(
                lambda value: {
                    'content': content_srv_tmpl.substitute(
                        priority=value.priority,
                        weight=value.weight,
                        port=value.port,
                        target=value.target,
                    )
                },
                record.values,
            )
        )
    elif record._type == "SSHFP":
        rrset_records = list(
            map(
                lambda value: {
                    'content': content_sshfp_tmpl.substitute(
                        algorithm=value.algorithm,
                        fingerprint_type=value.fingerprint_type,
                        fingerprint=value.fingerprint,
                    )
                },
                record.values,
            )
        )
    else:
        raise SelectelException(
            f'DNS Record with type: {record._type} not supported'
        )
    rrset["records"] = rrset_records
    return rrset


def to_octodns_record_data(rrset):
    rrset_type = rrset["type"]
    octodns_record = dict(type=rrset_type, ttl=rrset["ttl"])
    record_values = []
    key_for_record_values = "values"
    if rrset_type in {"A", "AAAA", "NS"}:
        record_values = [r['content'] for r in rrset["records"]]
    elif rrset_type in {"CNAME", "ALIAS"}:
        key_for_record_values = "value"
        record_values = rrset["records"][0]["content"]
    elif rrset_type == "TXT":
        record_values = [r['content'].strip('"\'') for r in rrset["records"]]
    elif rrset_type == "MX":
        for record in rrset["records"]:
            preference, exchange = record["content"].split(" ")
            record_values.append(
                {'preference': preference, 'exchange': exchange}
            )
    elif rrset_type == "SRV":
        for record in rrset["records"]:
            priority, weight, port, target = record["content"].split(" ")
            record_values.append(
                {
                    'priority': priority,
                    'weight': weight,
                    'port': port,
                    'target': target,
                }
            )
    elif rrset_type == "SSHFP":
        for record in rrset["records"]:
            algorithm, fingerprint_type, fingerprint = record["content"].split(
                " "
            )
            record_values.append(
                {
                    'algorithm': algorithm,
                    'fingerprint_type': fingerprint_type,
                    'fingerprint': fingerprint,
                }
            )
    else:
        raise SelectelException(
            f'DNS Record with type: {rrset_type} not supported'
        )
    octodns_record[key_for_record_values] = record_values
    return octodns_record



#
#
#

from logging import getLogger

from octodns.idna import idna_decode
from octodns.provider.base import BaseProvider
from octodns.record import Record, SshfpRecord, Update

from octodns_selectel.version import __version__ as provider_version

from .dns_client import DNSClient
from .exceptions import ApiException
from .mappings import to_octodns_record_data, to_selectel_rrset


class SelectelProvider(BaseProvider):
    SUPPORTS_GEO = False
    SUPPORTS = set(
        ('A', 'AAAA', 'ALIAS', 'CNAME', 'MX', 'NS', 'TXT', 'SRV', 'SSHFP')
    )
    MIN_TTL = 60

    def __init__(self, id, token, *args, **kwargs):
        self.log = getLogger(f'SelectelProvider[{id}]')
        self.log.debug('__init__: id=%s', id)
        super().__init__(id, *args, **kwargs)
        self._client = DNSClient(provider_version, token)
        self._zones = self.group_existing_zones_by_name()
        self._zone_rrsets = {}

    def _include_change(self, change):
        if isinstance(change, Update):
            existing = change.existing.data
            new = change.new.data
            new['ttl'] = max(self.MIN_TTL, new['ttl'])
            if isinstance(change.new, SshfpRecord):
                for i in range(0, len(change.new.rr_values)):
                    change.new.rr_values[i].fingerprint = change.new.rr_values[
                        i
                    ].fingerprint.lower()
            if new == existing:
                self.log.debug(
                    '_include_changes: new=%s, found existing=%s', new, existing
                )
                return False
        return True

    def _apply(self, plan):
        desired = plan.desired
        changes = plan.changes
        zone_name = idna_decode(desired.name)
        self.log.debug(
            '_apply: zone=%s, len(changes)=%d', zone_name, len(changes)
        )
        if not self._is_zone_already_created(zone_name):
            self.create_zone(zone_name)
        zone_id = self._get_zone_id_by_name(zone_name)
        for change in changes:
            action = change.__class__.__name__.lower()
            if action == 'create':
                self._apply_create(zone_id, change)
            if action == 'update':
                self._apply_update(zone_id, change)
            if action == 'delete':
                self._apply_delete(zone_id, change)

    def _is_zone_already_created(self, zone_name):
        return zone_name in self._zones.keys()

    def _get_rrset_id(self, zone_name, rrset_type, rrset_name):
        return next(
            filter(
                lambda rrset: rrset["type"] == rrset_type
                and rrset["name"] == rrset_name,
                self._zone_rrsets[zone_name],
            )
        )["id"]

    def _apply_create(self, zone_id, change):
        new_record = change.new
        rrset = to_selectel_rrset(new_record)
        self.create_rrset(zone_id, rrset)

    def _apply_update(self, zone_id, change):
        existing = change.existing
        rrset_id = self._get_rrset_id(
            idna_decode(existing.zone.name),
            existing._type,
            idna_decode(existing.fqdn),
        )
        data_for_update = to_selectel_rrset(change.new)
        self.update_rrset(zone_id, rrset_id, data_for_update)

    def _apply_delete(self, zone_id, change):
        existing = change.existing
        rrset_id = self._get_rrset_id(
            idna_decode(existing.zone.name),
            existing._type,
            idna_decode(existing.fqdn),
        )
        self.delete_rrset(zone_id, rrset_id)

    def populate(self, zone, target=False, lenient=False):
        zone_name = idna_decode(zone.name)
        self.log.debug(
            'populate: name=%s, target=%s, lenient=%s',
            zone_name,
            target,
            lenient,
        )
        before = len(zone.records)
        rrsets = []
        if self._is_zone_already_created(zone_name):
            rrsets = self.list_rrsets(zone)
        for rrset in rrsets:
            rrset_type = rrset['type']
            if rrset_type in self.SUPPORTS:
                record_data = to_octodns_record_data(rrset)
                rrset_hostname = zone.hostname_from_fqdn(rrset['name'])
                record = Record.new(
                    zone,
                    rrset_hostname,
                    record_data,
                    source=self,
                    lenient=lenient,
                )
                zone.add_record(record)
        self.log.info('populate: found %s records', len(zone.records) - before)

    def _get_zone_id_by_name(self, zone_name):
        return self._zones.get(zone_name, False)["id"]

    def create_zone(self, name):
        self.log.debug('Create zone: %s', name)
        zone = self._client.create_zone(name)
        self._zones[zone["name"]] = zone
        return zone

    def list_zones(self):
        # This method is called dynamically in octodns.Manager._preprocess_zones()
        # and required for use of "*" if provider is source.
        return [zone_name for zone_name in self._zones]

    def group_existing_zones_by_name(self):
        self.log.debug('View zones')
        return {zone['name']: zone for zone in self._client.list_zones()}

    def list_rrsets(self, zone):
        zone_name = idna_decode(zone.name)
        self.log.debug('View rrsets. Zone: %s', zone_name)
        zone_id = self._get_zone_id_by_name(zone_name)
        zone_rrsets = self._client.list_rrsets(zone_id)
        self._zone_rrsets[zone_name] = zone_rrsets
        return zone_rrsets

    def create_rrset(self, zone_id, data):
        self.log.debug('Create rrset. Zone id: %s, data %s', zone_id, data)
        return self._client.create_rrset(zone_id, data)

    def update_rrset(self, zone_id, rrset_id, data):
        self.log.debug(
            f'Update rrsets. Zone id: {zone_id}, rrset id: {rrset_id}'
        )
        try:
            self._client.update_rrset(zone_id, rrset_id, data)
        except ApiException as api_exception:
            self.log.warning(
                f'Failed to update rrset {rrset_id}. {api_exception}'
            )

    def delete_rrset(self, zone_id, rrset_id):
        self.log.debug(
            f'Delete rrsets. Zone id: {zone_id}, rrset id: {rrset_id}'
        )
        try:
            self._client.delete_rrset(zone_id, rrset_id)
        except ApiException as api_exception:
            self.log.warning(
                f'Failed to delete rrset {rrset_id}. {api_exception}'
            )



from collections import defaultdict
from logging import getLogger

from requests import Session
from requests.exceptions import HTTPError

from octodns import __version__ as octodns_version
from octodns.provider import ProviderException
from octodns.provider.base import BaseProvider
from octodns.record import Record, Update

from octodns_selectel.version import __version__ as provider_version


def require_root_domain(fqdn):
    if fqdn.endswith('.'):
        return fqdn

    return f'{fqdn}.'


class SelectelAuthenticationRequired(ProviderException):
    def __init__(self, msg):
        message = 'Authorization failed. Invalid or empty token.'
        super().__init__(message)


class SelectelProvider(BaseProvider):
    SUPPORTS_GEO = False

    SUPPORTS = set(
        ('A', 'AAAA', 'ALIAS', 'CNAME', 'MX', 'NS', 'TXT', 'SRV', 'SSHFP')
    )

    MIN_TTL = 60

    PAGINATION_LIMIT = 50

    API_URL = 'https://api.selectel.ru/domains/v1'

    def __init__(self, id, token, *args, **kwargs):
        self.log = getLogger(f'SelectelProvider[{id}]')
        self.log.debug('__init__: id=%s', id)
        super().__init__(id, *args, **kwargs)

        self._sess = Session()
        self._sess.headers.update(
            {
                'X-Token': token,
                'Content-Type': 'application/json',
                'User-Agent': f'octodns/{octodns_version} octodns-selectel/{provider_version}',
            }
        )
        self._zone_records = {}
        self._domain_list = self.domain_list()
        self._zones = None

    def _request(self, method, path, params=None, data=None):
        self.log.debug('_request: method=%s, path=%s', method, path)

        url = f'{self.API_URL}{path}'
        resp = self._sess.request(method, url, params=params, json=data)

        self.log.debug('_request: status=%s', resp.status_code)
        if resp.status_code == 401:
            raise SelectelAuthenticationRequired(resp.text)
        elif resp.status_code == 404:
            return {}
        resp.raise_for_status()
        if method == 'DELETE':
            return {}
        return resp.json()

    def _get_total_count(self, path):
        url = f'{self.API_URL}{path}'
        resp = self._sess.request('HEAD', url)
        return int(resp.headers['X-Total-Count'])

    def _request_with_pagination(self, path, total_count):
        result = []
        for offset in range(0, total_count, self.PAGINATION_LIMIT):
            result += self._request(
                'GET',
                path,
                params={'limit': self.PAGINATION_LIMIT, 'offset': offset},
            )
        return result

    def _include_change(self, change):
        if isinstance(change, Update):
            existing = change.existing.data
            new = change.new.data
            new['ttl'] = max(self.MIN_TTL, new['ttl'])
            if new == existing:
                self.log.debug(
                    '_include_changes: new=%s, found existing=%s', new, existing
                )
                return False
        return True

    def _apply(self, plan):
        desired = plan.desired
        changes = plan.changes
        self.log.debug(
            '_apply: zone=%s, len(changes)=%d', desired.name, len(changes)
        )

        zone_name = desired.name[:-1]
        for change in changes:
            class_name = change.__class__.__name__
            getattr(self, f'_apply_{class_name}'.lower())(zone_name, change)

    def _apply_create(self, zone_name, change):
        new = change.new
        params_for = getattr(self, f'_params_for_{new._type}')
        for params in params_for(new):
            self.create_record(zone_name, params)

    def _apply_update(self, zone_name, change):
        self._apply_delete(zone_name, change)
        self._apply_create(zone_name, change)

    def _apply_delete(self, zone_name, change):
        existing = change.existing
        self.delete_record(zone_name, existing._type, existing.name)

    def list_zones(self):
        # This method is called dynamically in octodns.Manager._preprocess_zones()
        # and required for use of "*" if provider is source.
        zones_without_dot = self.domain_list()
        return [
            require_root_domain(zone_name) for zone_name in zones_without_dot
        ]

    def _params_for_multiple(self, record):
        for value in record.values:
            yield {
                'content': value,
                'name': record.fqdn,
                'ttl': max(self.MIN_TTL, record.ttl),
                'type': record._type,
            }

    def _params_for_single(self, record):
        yield {
            'content': record.value,
            'name': record.fqdn,
            'ttl': max(self.MIN_TTL, record.ttl),
            'type': record._type,
        }

    def _params_for_MX(self, record):
        for value in record.values:
            yield {
                'content': value.exchange,
                'name': record.fqdn,
                'ttl': max(self.MIN_TTL, record.ttl),
                'type': record._type,
                'priority': value.preference,
            }

    def _params_for_SRV(self, record):
        for value in record.values:
            yield {
                'name': record.fqdn,
                'target': value.target,
                'ttl': max(self.MIN_TTL, record.ttl),
                'type': record._type,
                'port': value.port,
                'weight': value.weight,
                'priority': value.priority,
            }

    def _params_for_SSHFP(self, record):
        for value in record.values:
            yield {
                'name': record.fqdn,
                'ttl': max(self.MIN_TTL, record.ttl),
                'type': record._type,
                'algorithm': value.algorithm,
                'fingerprint_type': value.fingerprint_type,
                'fingerprint': value.fingerprint,
            }

    _params_for_A = _params_for_multiple
    _params_for_AAAA = _params_for_multiple
    _params_for_NS = _params_for_multiple
    _params_for_TXT = _params_for_multiple

    _params_for_CNAME = _params_for_single
    _params_for_ALIAS = _params_for_single

    def _data_for_A(self, _type, records):
        return {
            'ttl': records[0]['ttl'],
            'type': _type,
            'values': [r['content'] for r in records],
        }

    _data_for_AAAA = _data_for_A

    def _data_for_NS(self, _type, records):
        return {
            'ttl': records[0]['ttl'],
            'type': _type,
            'values': [require_root_domain(r["content"]) for r in records],
        }

    def _data_for_MX(self, _type, records):
        values = []
        for record in records:
            values.append(
                {
                    'preference': record['priority'],
                    'exchange': require_root_domain(record["content"]),
                }
            )
        return {'ttl': records[0]['ttl'], 'type': _type, 'values': values}

    def _data_for_CNAME(self, _type, records):
        only = records[0]
        return {
            'ttl': only['ttl'],
            'type': _type,
            'value': require_root_domain(only["content"]),
        }

    _data_for_ALIAS = _data_for_CNAME

    def _data_for_TXT(self, _type, records):
        return {
            'ttl': records[0]['ttl'],
            'type': _type,
            'values': [r['content'] for r in records],
        }

    def _data_for_SRV(self, _type, records):
        values = []
        for record in records:
            values.append(
                {
                    'priority': record['priority'],
                    'weight': record['weight'],
                    'port': record['port'],
                    'target': require_root_domain(record["target"]),
                }
            )

        return {'type': _type, 'ttl': records[0]['ttl'], 'values': values}

    def _data_for_SSHFP(self, _type, records):
        values = []
        for record in records:
            values.append(
                {
                    'algorithm': record['algorithm'],
                    'fingerprint_type': record['fingerprint_type'],
                    'fingerprint': f'{record["fingerprint"]}',
                }
            )

        return {'type': _type, 'ttl': records[0]['ttl'], 'values': values}

    def populate(self, zone, target=False, lenient=False):
        self.log.debug(
            'populate: name=%s, target=%s, lenient=%s',
            zone.name,
            target,
            lenient,
        )
        before = len(zone.records)
        records = self.zone_records(zone)
        if records:
            values = defaultdict(lambda: defaultdict(list))
            for record in records:
                name = zone.hostname_from_fqdn(record['name'])
                _type = record['type']
                if _type in self.SUPPORTS:
                    values[name][record['type']].append(record)
            for name, types in values.items():
                for _type, records in types.items():
                    data_for = getattr(self, f'_data_for_{_type}')
                    data = data_for(_type, records)
                    record = Record.new(
                        zone, name, data, source=self, lenient=lenient
                    )
                    zone.add_record(record)
        self.log.info(
            'populate:   found %s records', len(zone.records) - before
        )

    def domain_list(self):
        path = '/'
        domains = {}
        domains_list = []

        total_count = self._get_total_count(path)
        domains_list = self._request_with_pagination(path, total_count)

        for domain in domains_list:
            domains[domain['name']] = domain
        return domains

    def zone_records(self, zone):
        path = f'/{zone.name[:-1]}/records/'
        zone_records = []

        total_count = self._get_total_count(path)
        zone_records = self._request_with_pagination(path, total_count)

        self._zone_records[zone.name] = zone_records
        return self._zone_records[zone.name]

    def create_domain(self, name, zone=""):
        path = '/'

        data = {'name': name, 'bind_zone': zone}

        resp = self._request('POST', path, data=data)
        self._domain_list[name] = resp
        return resp

    def create_record(self, zone_name, data):
        self.log.debug('Create record. Zone: %s, data %s', zone_name, data)
        if zone_name in self._domain_list.keys():
            domain_id = self._domain_list[zone_name]['id']
        else:
            domain_id = self.create_domain(zone_name)['id']

        path = f'/{domain_id}/records/'
        return self._request('POST', path, data=data)

    def delete_record(self, domain, _type, zone):
        self.log.debug('Delete records. Domain: %s, Type: %s', domain, _type)
        domain_id = self._domain_list[domain]['id']
        records = self._zone_records.get(f'{domain}.', False)
        if not records:
            path = f'/{domain_id}/records/'
            records = self._request('GET', path)

        full_domain = f'{zone}.{domain}' if zone else domain
        delete_count, skip_count = 0, 0
        for record in records:
            if record['type'] == _type and record['name'] == full_domain:
                record_id = record["id"]
                path = f'/{domain_id}/records/{record_id}'
                try:
                    self._request('DELETE', path)
                    delete_count += 1
                except HTTPError:
                    skip_count += 1
                    self.log.warning(f'Failed to delete record {record_id}')

        self.log.debug(
            f'Deleted {delete_count} records. Skipped {skip_count} records'
        )
